﻿<#
.SYNOPSIS 
Скрипт автообновления Снегопата 2.0
	
.DESCRIPTION
Скрипт обновляет и при необходимости создает с нуля и клонирует
тестовый репозитарий Снегопата, а также пользовательские репозитарии fossil
	
.EXAMPLE
Get-Latest20

.EXAMPLE
Get-Latest20 -Path "C:\Snegopat20"
		
.LINK
https://snegopat.ru/forum/viewtopic.php?f=3&t=662
		
.NOTES
Версия 2.0.0.4

#>

param(
	[parameter(Mandatory=$False, HelpMessage="Задает каталог в котором будут размещаться содержимое репозитария.")]
	[string]$Path = (Get-Location).path,
	[parameter(Mandatory=$False, HelpMessage="Задает каталог в котором будут размещаться файлы метаданных репозитария.")]
	[string]$RepoPath = "$Path\repo",
	[parameter(Mandatory=$False, HelpMessage="Указывет расположение файла fossil.exe.")]
	[string]$Fossil = (Join-Path -Path $Path -ChildPath "fossil.exe"),
	[parameter(Mandatory=$False, HelpMessage="Указывет расположение ini файла с пользовательскими репозиториями.")]
	[string]$IniPath = (Join-Path -Path $Path -ChildPath "Get-Latest.ini")
)

#Region Init

Clear-Host

Write-Debug -Message "Argument Path=$Path"
Write-Debug -Message "Argument RepoPath=$RepoPath"
Write-Debug -Message "Argument Fossil=$Fossil"
Write-Debug -Message "Argument IniPath=$IniPath"

if (-not (Test-Path -LiteralPath $Path -PathType Container)) { 
	Throw "Каталог $Path не найден. Скрипт остановлен!" 
}

if (-not (Test-Path -LiteralPath $Fossil -PathType Leaf)) { 
	Throw "Не найден fossil.exe. Скрипт остановлен!" 
} else {
	Write-Debug -Message "fossil.exe найден: $Fossil"
}

#EndRegion

#region Functions

Function New-Repository {
	[CmdletBinding()]
	[OutputType([hashtable])]
	Param(
		[parameter(Mandatory=$true)]
		[alias("Key")]
		[string]$Name, 
		[parameter(Mandatory=$true)]
		[string]$Description,
		[parameter(Mandatory=$true)]
		[string]$URLPath="" ,
		[parameter(Mandatory=$false)]
		[string]$Path="",
		[parameter(Mandatory=$false)]
		[string]$RepoPath="" 
	)
		
	if ($Path -eq "")     { $Path     = "$($script:Path)\$name" }
	if ($RepoPath -eq "") { $RepoPath = "$($script:RepoPath)\$name.fossil" }
	
	Write-Output @{
		Name        = $Name
		Description = $Description
		Path		= $Path
		URLPath     = $URLPath
		RepoPath    = $RepoPath
		}
}

Function Add-Repositories {
	[CmdletBinding()]
	[OutputType([void])]
	param(
		[parameter(Mandatory=$true)]
		[string]$Path,
		[parameter(Mandatory=$true)]
		[hashtable[]]$Repositories
	)
	
	[hashtable]$Repository = @{}
    switch -regex -file $Path {
        "^\[(?<name>.+)\]" {
			if($Repository.Count -ne 0) { $Repositories += New-Repository @Repository }
            $Repository.Clear()
			$Repository.name = $matches.name
        }
        "(?<key>.+?)\s*=(?<value>.*)" { $Repository[$matches.key] = $matches.value }
    }
	
	if($Repository.Count -ne 0) { 
		$Repositories += New-Repository @Repository 
	}
}



Function Run-Fossil {
	[CmdletBinding()]
	[OutputType([boolean])]
	Param(
		[parameter(Mandatory=$true)]
		[string]$Command,
		[parameter(Mandatory=$false)]
		[string[]]$Arguments=@(),
		[parameter(Mandatory=$false)]
		[string]$Path = $script:Path
	)

	[string[]]$ArgumentList = @($Command) + $Arguments
	
	Write-Debug -Message "$($ArgumentList)"

	Push-Location
	Set-Location -Path $Path
	$process = Start-Process -FilePath $script:Fossil -ArgumentList $ArgumentList -Wait -NoNewWindow -PassThru
	Pop-Location
	
	Write-Output ($process.ExitCode -eq 0)
}

Function New-Folder {
	[CmdletBinding()]
	[OutputType([void])]
	Param(
		[parameter(Mandatory=$true)]
		[string]$Path
	)

	If(-not (Test-Path -LiteralPath $Path -PathType Container)) { 
		New-Item -ItemType Directory -Path $Path -ErrorAction Stop | Out-Null 
	}
}

Function Test-Repository {
	[CmdletBinding()]
	[OutputType([hashtable],[void])]
	Param(
		[parameter(Mandatory=$true,ValueFromPipeline=$true)]
		[hashtable]$Repository,
		[switch]$PassThru
	)

	process {
		New-Folder -Path $Repository.Path
		if($PassThru) { Write-Output $Repository }
	}
}

Function Clone-Repository {
	[CmdletBinding()]
	[OutputType([hashtable],[void])]
	Param(
		[parameter(Mandatory=$true,ValueFromPipeline=$true)]
		[hashtable]$Repository,
		[switch]$PassThru
	)
	
	begin {
		[hashtable]$Logins = @{}
	}
	
	process {
		
		Function Read-Login {
			[CmdletBinding()]
			[OutputType([string])]
			Param([string]$Domain)
			
			[string]$Login = ""
			if($Domain -ne "") {
				[string]$hashedLogin = $Logins[$Domain]
				if($hashedLogin -eq "" ) {
					$hashedLogin = Read-Host -Prompt "Укажите логин на $Domain"
					$Logins[$Domain] = $hashedLogin
				}
				$Login = $hashedLogin
			}	
		
			If ($Login -eq "") { Throw "Не указан логин" }
			Write-Output $Login
		}
		
		If (-not (Test-Path -LiteralPath $Repository.RepoPath)) {

			if ($Repository.URLPath -match "(?<protocol>https?:\/\/)(?<domain>[^\/\s]+)(?<path>.*)") {
				
				[string]$URLProtocol = $matches.protocol
				[string]$URLDomain   = $matches.domain
				[string]$URLPath     = $matches.path
				[string]$Login = Read-Login -Domain $URLDomain
				
				Write-Host ""
				Write-Host "Клонирование" $Repository.Description
				Write-Host ""		
				
			}
			
			[string[]]$Arguments = @(
				"$URLProtocol$Login@$URLDomain$URLPath", 
				"-A $Login", 
				"""$($Repository.RepoPath)"""
				)		
			[boolean]$ExitStatus = Run-Fossil -Command clone -Arguments $Arguments
			If ($ExitStatus -eq $false) { Throw "Не удалось клонировать $($Repository.Description)"	}
		}
	
		if($PassThru) { Write-Output $Repository }
	}
}

Function Open-Repository {
	[CmdletBinding()]
	[OutputType([hashtable],[void])]
	Param(
		[parameter(Mandatory=$true,ValueFromPipeline=$true)]
		[hashtable]$Repository,
		[switch]$PassThru
	)
	
	process {
		
		If (-not (Test-Path -LiteralPath "$($Repository.Path)\_fossil_")) {
			[boolean]$ExitStatus = Run-Fossil -Command open -Arguments """$($Repository.RepoPath)""" -Path $Repository.Path
			If ($ExitStatus -eq $false) { Throw "Не удалось открыть $($Repository.Description)" }
		}

		if($PassThru) { Write-Output $Repository }
	}
}

Function Update-Repository {
	[CmdletBinding()]
	[OutputType([hashtable],[void])]
	Param(
		[parameter(Mandatory=$true,ValueFromPipeline=$true)]
		[hashtable]$Repository,
		[switch]$PassThru
	)
	
	process {
		If (Test-Path -LiteralPath "$($Repository.Path)\_fossil_") {
			Run-Fossil -Command pull -Path $Repository.Path | Out-Null
			Run-Fossil -Command update -Path $Repository.Path | Out-Null
		}
		if($PassThru) { Write-Output $Repository }
	}
}

Function Close-Repository {
	[CmdletBinding()]
	[OutputType([hashtable],[void])]
	Param(
		[parameter(Mandatory=$true,ValueFromPipeline=$true)]
		[hashtable]$Repository,
		[switch]$PassThru
	)
	
	process {
		
		If (Test-Path -LiteralPath "$($Repository.Path)\_fossil_") {
			Run-Fossil -Command close -Path $Repository.Path | Out-Null
		}
		if($PassThru) { Write-Output $Repository }
	}

	end {
	}
}

#EndRegion



New-Folder -Path $RepoPath

[hashtable[]]$Repositories = @()
$Repositories += New-Repository -Name "core" -URLPath "http://snegopat.ru:9004" -Description "репозиторий Снегопата 2.0" -Path $Path

if (Test-Path -Path $IniPath -PathType Leaf) {
	Add-Repositories -Path $IniPath -Repositories $Repositories
}

$Repositories |
	Test-Repository	-PassThru 	| 
	Clone-Repository -PassThru | 
	Open-Repository -PassThru | 
	Update-Repository -PassThru | 
	Close-Repository